package ninja.thefirearchmage.games.fourtykcalculator;

import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class AttackScenario {
    private static final int PROBABILITY_DENOMINATOR_D6 = 6;

    // Attack attributes
    private UnitAttackModifiers unitAttackModifiers;
    private List<Tuple2<Weapon, Integer>> weapons;
    private boolean unmodifiableHit;
    private boolean attackerWasStationary;
    private boolean attackerCharged;
    private boolean attackerInRapidFireRange;
    private boolean hasHitRollBonus;
    private boolean hasWoundRollBonus;
    private boolean hasLineOfSight;
    private boolean hasRerollHitsOnes;
    private boolean hasRerollHitsFailures;
    private boolean hasRerollHitsAny;
    private boolean hasRerollWoundsOnes;
    private boolean hasRerollWoundsFailures;
    private boolean hasRerollWoundsAny;
    private boolean fishCriticalHits;
    private boolean fishCriticalWounds;
    private int criticalHitValue;
    private int criticalWoundValue;

    // Defense attributes
    private DefensiveProfile defensiveProfile;
    private Optional<Reroll> normalSaveReroll;
    private boolean hasHitPenalty;
    private boolean hasWoundRollPenalty;
    private boolean isInCover;
    private boolean woundRollPenaltyIfStrengthIsHigher;
    private int damageReduction;

    // Result
    private AttackScenarioResolution results;

    public AttackScenario() {
        weapons = new ArrayList<>();
        unmodifiableHit = false;
        attackerWasStationary = false;
        attackerCharged = true;
        attackerInRapidFireRange = true;
        hasHitRollBonus = false;
        hasWoundRollBonus = false;
        hasLineOfSight = true;
        hasRerollHitsOnes = false;
        hasRerollHitsFailures = false;
        hasRerollHitsAny = false;
        hasRerollWoundsOnes = false;
        hasRerollWoundsFailures = false;
        hasRerollWoundsAny = false;
        fishCriticalHits = false;
        fishCriticalWounds = false;
        criticalHitValue = 6;
        criticalWoundValue = 6;

        defensiveProfile = null;
        normalSaveReroll = Optional.empty();
        hasHitPenalty = false;
        hasWoundRollPenalty = false;
        isInCover = false;
        woundRollPenaltyIfStrengthIsHigher = false;
        damageReduction = 0;

        results = new AttackScenarioResolution();
    }

    public AttackScenario addWeapon(Weapon weaponType, int numberOfWeaponsInUnit) {
        weapons.add(Tuple.of(weaponType, numberOfWeaponsInUnit));
        return this;
    }

    public AttackScenario withDefensiveProfile(DefensiveProfile defensiveProfile) {
        this.defensiveProfile = defensiveProfile;
        return this;
    }

    public AttackScenarioResolution resolveAttack() {
        // Filter out weapons that cannot participate in this attack
        var validWeapons = weapons.stream().filter( w -> hasLineOfSight || w._1.hasIndirectFire())
                .toList();

        // Attack step
        var totalDamage = 0d;
        var totalPreventedDamage = 0d;
        for(var weaponData : validWeapons) {
            Weapon weapon = weaponData._1;
            int amountOfWeapons = weaponData._2;

            weapon.generateAttackResolution(defensiveProfile, )
            var weaponAttackResolution = new WeaponAttackResolution();

            // Attack funnel
            var passingHits = calculatePassingHits(weapon, amountOfWeapons);
            var passingWounds = calculatePassingWounds(weapon, amountOfWeapons, passingHits);
            var normalWounds = passingWounds._1;
            var mortalWounds = passingWounds._2;

            // Defence funnel
            var savesResolution = defensiveProfile.saveWounds(weapon, normalWounds, mortalWounds);
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_NORMAL, normalWounds - savesResolution.getPreventedNormalWounds());
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_MORTAL, mortalWounds - savesResolution.getPreventedMortalWoundsFromFNP());
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_TOTAL, normalWounds + mortalWounds - savesResolution.getPreventedWounds());

            totalDamage += normalWounds + mortalWounds - savesResolution.getPreventedWounds();
            totalPreventedDamage += savesResolution.getPreventedWounds();
        }
        results.setTotalPreventedDamage(totalDamage);
        results.setTotalPreventedDamage(totalPreventedDamage);

        return results;
    }

    /**
     * Returns a tuple of normal passing wounds and mortal wounds so that the caller can handle them differently.
     */
    private Tuple2<Double, Double> calculatePassingWounds(Weapon weapon, int amountOfWeapons, double passingHits) {
        double lethalHitsCount = calculateLethalHits(weapon, amountOfWeapons);
        var normalHits = passingHits - lethalHitsCount;
        results.setWeaponStat(weapon, WeaponStat.NON_LETHAL_HITS_COUNT, normalHits);

        int woundRollGoal = calculateWoundRollGoal(weapon);
        int woundSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(woundRollGoal);

        var passingNormalWounds = normalHits * ( (double) woundSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
        results.setWeaponStat(weapon, WeaponStat.WOUNDS_COUNT_WITHOUT_REROLL, passingNormalWounds+lethalHitsCount);
        var woundsRerolls = prepareWoundRerolls(woundRollGoal);
        if(woundsRerolls.isPresent()) {
            var rerolls = woundsRerolls.get();
            var failedWounds = normalHits * ((double) (1 - woundSuccessProbabilityNumerator) / PROBABILITY_DENOMINATOR_D6);
            var extraWoundsFromRerolls = failedWounds * rerolls.getProbabilityNumerator() / PROBABILITY_DENOMINATOR_D6;
            results.setWeaponStat(weapon, WeaponStat.WOUNDS_COUNT_FROM_REROLL, extraWoundsFromRerolls);
            passingNormalWounds += extraWoundsFromRerolls;
        }

        // Mortal wounds
        var passingMortalWounds = 0d;
        if(weapon.hasDevastatingWounds()) {
            var criticalWoundRollGoal = weapon.getBestAntiFor(defensiveProfile)
                    .orElse(criticalWoundValue);
            var criticalProbability = calculateProbabilityNumeratorFrom(criticalWoundRollGoal);
            passingMortalWounds += (normalHits*criticalProbability/PROBABILITY_DENOMINATOR_D6);
            if(fishCriticalWounds && hasRerollWoundsAny) {
                passingMortalWounds += (normalHits * (1-criticalProbability)/PROBABILITY_DENOMINATOR_D6)*criticalProbability/PROBABILITY_DENOMINATOR_D6;
            }

            results.setWeaponStat(weapon, WeaponStat.MORTAL_WOUNDS_COUNT, passingMortalWounds);
            // passing wounds was calculated from the whole, but we must substract the critical wounds from the whole
            // to be able to handle them separately, as mortal wounds replace normal wounds.
            passingNormalWounds -= passingMortalWounds;
        }

        passingNormalWounds += lethalHitsCount;
        results.setWeaponStat(weapon, WeaponStat.WOUNDS_TOTAL_COUNT, passingNormalWounds+passingMortalWounds);

        return Tuple.of(passingNormalWounds, passingMortalWounds);
    }

    private double calculateLethalHits(Weapon weapon, int amountOfWeapons) {
        var lethalHitsCount = 0d;

        if(weapon.hasLethalHits()) {
            double weaponAttacks = calculateBasicWeaponAttacks(weapon, amountOfWeapons);
            var lethalHitsProbabilityNumerator = calculateProbabilityNumeratorFrom(criticalHitValue);

            lethalHitsCount = weaponAttacks * ((double) lethalHitsProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
            results.setWeaponStat(weapon, WeaponStat.LETHAL_HITS_COUNT, lethalHitsCount);
        }

        return lethalHitsCount;
    }


    private double calculatePassingHits(Weapon weapon, int amountOfWeapons) {
        int hitRollGoal = calculateHitRollGoal(weapon);
        int hitSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(hitRollGoal);
        double weaponAttacks = calculateWeaponAttacks(weapon, amountOfWeapons);

        // Base passing hits without rerolls
        double passingHits = (weaponAttacks*amountOfWeapons) * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
        results.setWeaponStat(weapon, WeaponStat.HITS_COUNT_WITHOUT_REROLL, passingHits);

        // Add hits from rerolls
        var hitRerolls = prepareHitRerolls(hitRollGoal);
        if(hitRerolls.isPresent()) {
            var rerolls = hitRerolls.get();
            var successfulHitFromRerolls = rerolls.getProbabilityNumerator() * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
            results.setWeaponStat(weapon, WeaponStat.HITS_COUNT_FROM_REROLL, successfulHitFromRerolls);
            passingHits += successfulHitFromRerolls;
        }
        results.setWeaponStat(weapon, WeaponStat.HITS_TOTAL_COUNT, passingHits);

        return passingHits;
    }

    private Optional<Reroll> prepareHitRerolls(int hitRollGoal) {
        if(hasRerollHitsOnes) {
            return Optional.of(Reroll.rerollOnOnes());
        }
        if(hasRerollHitsFailures || hasRerollHitsAny) {
            return Optional.of(Reroll.rerollFailures(hitRollGoal));
        }

        return Optional.empty();
    }

    private Optional<Reroll> prepareWoundRerolls(int woundRollGoal) {
        if(hasRerollWoundsOnes) {
            return Optional.of(Reroll.rerollOnOnes());
        }
        if(hasRerollWoundsFailures || hasRerollWoundsAny) {
            return Optional.of(Reroll.rerollFailures(woundRollGoal));
        }

        return Optional.empty();
    }

    private double calculateWeaponAttacks(Weapon weapon, int amountOfWeapons) {
        double weaponAttacks = calculateBasicWeaponAttacks(weapon, amountOfWeapons);
        results.setWeaponStat(weapon, WeaponStat.POTENTIAL_HITS_COUNT, weaponAttacks);

        // Add sustained hits
        if(weapon.getSustainedHits().isPresent()) {
            var sustainedHitsValue = weapon.getSustainedHits().get();

            double nonSustainedProbability = (double) (6 - (7 - criticalHitValue)) /PROBABILITY_DENOMINATOR_D6;
            double sustainedProbability = (double) (7 - criticalHitValue) /PROBABILITY_DENOMINATOR_D6;
            if(fishCriticalHits && hasRerollHitsAny) {
                sustainedProbability += nonSustainedProbability*sustainedProbability;
            }
            var sustainedHitsAmount = weaponAttacks * sustainedProbability * sustainedHitsValue;
            results.setWeaponStat(weapon, WeaponStat.SUSTAINED_HITS_COUNT, sustainedHitsAmount);
            weaponAttacks += sustainedHitsAmount;
        }

        return weaponAttacks;
    }

    private double calculateBasicWeaponAttacks(Weapon weapon, int amountOfWeapons) {
        double weaponAttacks = weapon.getAttacks();
        if(weapon.getRapidFire().isPresent() && attackerInRapidFireRange) {
            weaponAttacks += weapon.getRapidFire().get();
        }
        weaponAttacks *= amountOfWeapons;

        return weaponAttacks;
    }

    private int calculateHitRollGoal(Weapon weapon) {
        var finalHitRollGoal = weapon.getHitValue();
        if(hasHitRollBonus) {
            finalHitRollGoal--;
        }
        if(hasHitPenalty && !unmodifiableHit) {
            finalHitRollGoal++;
        }
        if(!hasLineOfSight && weapon.hasIndirectFire()) {
            finalHitRollGoal++;
            finalHitRollGoal = Math.max(finalHitRollGoal, 4);
        }

        return finalHitRollGoal;
    }

    /**
     * Converts from a roll goal to a probability denominator of a fraction P/6. For hit and wound rolls, the die used
     * is D6.
     */
    private static int calculateProbabilityNumeratorFrom(int rollGoal) {
        // Because a roll of 1 always fails and a roll of 6 always succeeds, we modify the roll goal to be between
        // values 2-5
        var modifiedRollGoal = Math.max(rollGoal, 2);

        return 7 - modifiedRollGoal;

    }

    private int calculateWoundRollGoal(Weapon weapon) {
        var woundRollGoal = 4;
        var finalWoundPenalty = hasWoundRollPenalty ? 1 : 0;
        var toughness = defensiveProfile.getToughness();
        var strength = weapon.getStrength();

        if(strength > toughness) {
            finalWoundPenalty += woundRollPenaltyIfStrengthIsHigher ? 1 : 0;
            woundRollGoal = strength >= toughness*2 ? 2 : 3;
        } else if(strength < toughness) {
            woundRollGoal = toughness*2 >= strength ? 6 : 5;
        }

        var finalWoundBonus = hasWoundRollBonus ? 1 : 0;
        if(attackerCharged && weapon.hasLance() && weapon.getRangeType() == WeaponRangeType.MELEE) {
            finalWoundBonus++;
        }

        // Clever way to set either -1, 0 or 1, as the wound roll can only be modified, in the end by 1
        var finalWoundRollModifier = Integer.compare(finalWoundBonus - finalWoundPenalty, 0);

        // Having a -1 to wound rolls implies that the roll goal moves by 1
        woundRollGoal -= finalWoundRollModifier;

        // Now apply any anti-X if it is better than the current wound roll
        var antiWoundRoll = weapon.getBestAntiFor(defensiveProfile);
        if(antiWoundRoll.isPresent()) {
            woundRollGoal = Math.min(woundRollGoal, antiWoundRoll.get());
        }

        return woundRollGoal;
    }
}
