package ninja.thefirearchmage.games.fourtykcalculator;

import lombok.Getter;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple3;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;

@Getter
public class DefensiveProfile {
    private final int toughness;
    private final int normalSave;
    private final int wounds;
    private final int bodies;
    private Optional<InvulnerableSave> invulnerableSave;
    private List<FeelNoPainEffect> feelNoPainEffects;
    private Set<UnitTypeKeyword> unitTypeKeywords;
    private boolean isInCover;
    private boolean isVisible;
    private boolean hasRerollOneNormalSaves;
    private boolean hasRerollFailedNormalSaves;
    private boolean hasRerollAnyNormalSaves;
    private boolean fishFor6NormalSaves;
    private boolean hasStealth;
    private Tuple3<Boolean, DamageReductionType, Integer> damageReductionModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> apReduction;

    public DefensiveProfile(int toughness, int wounds, int bodies, int normalSave) {
        this(toughness, wounds, bodies, normalSave, Optional.empty(), Optional.empty(),
                new HashSet<>(), false, false, false,
                false, true);
    }

    public DefensiveProfile(int toughness, int wounds, int bodies, int normalSave, Optional<InvulnerableSave> invulnerableSave,
                            Optional<FeelNoPainEffect> feelNoPain, Set<UnitTypeKeyword> unitTypeKeywords,
                            boolean isInCover, boolean hasRerollOneNormalSaves, boolean hasRerollFailedNormalSaves,
                            boolean hasRerollAnyNormalSaves, boolean isVisible) {
        this.toughness = toughness;
        this.normalSave = normalSave;
        this.wounds = wounds;
        this.bodies = bodies;
        this.invulnerableSave = invulnerableSave;
        this.feelNoPain = feelNoPain;
        this.unitTypeKeywords = unitTypeKeywords;
        this.isInCover = isInCover;
        this.hasRerollOneNormalSaves = hasRerollOneNormalSaves;
        this.hasRerollFailedNormalSaves = hasRerollFailedNormalSaves;
        this.hasRerollAnyNormalSaves = hasRerollAnyNormalSaves;
        this.isVisible = isVisible;
    }

    public DefensiveProfile withUnitTypes(Set<UnitTypeKeyword> types) {
        this.unitTypeKeywords.addAll(types);
        return this;
    }
    public DefensiveProfile withRerollsConfig(boolean rerollOneNormalSaves, boolean rerollFailedNormalSaves,
                                              boolean rerollAnyNormalSaves) {
        this.hasRerollOneNormalSaves = rerollOneNormalSaves;
        this.hasRerollFailedNormalSaves = rerollFailedNormalSaves;
        this.hasRerollAnyNormalSaves = rerollAnyNormalSaves;
        return this;
    }
    public DefensiveProfile withInvulnerabilitySave(InvulnerableSave invulnerabilitySave) {
        this.invulnerableSave = Optional.of(invulnerabilitySave);
        return this;
    }
    public DefensiveProfile withFeelNoPainConfig(FeelNoPainEffect feelNoPainConfig) {
        this.feelNoPain = Optional.of(feelNoPainConfig);
        return this;
    }
    public DefensiveProfile setInCover(boolean inCover) {
        this.isInCover = inCover;
        return this;
    }
    public DefensiveProfile setIsVisible(boolean isVisible) {
        this.isVisible = isVisible;
        return this;
    }

    /**
     * Returns a detailed description of how the wounds applied to this defensive profile were prevented
     */
    public SaveResolution saveWounds(Weapon weapon, double normalWounds, double mortalWounds) {
        var resolution = new SaveResolution();

        // Get the best save for this weapon
        var saveRollGoal = calculateBestSaveFor(weapon);
        var saveProbabilityNumerator = 7 - saveRollGoal;

        // Normal wounds
        var savedNormalWounds = normalWounds * saveProbabilityNumerator/6;
        resolution.setPreventedNormalWoundsFromSavesWithoutRerolls(savedNormalWounds);
        var saveRerolls = calculateBestReroll(saveRollGoal);
        if(saveRerolls.isPresent()) {
            var rerollProbability = saveRerolls.get().getProbabilityNumerator();
            var savedNormalWoundsFromRerolls = normalWounds*(1 - saveProbabilityNumerator)/6 * rerollProbability/6;
            resolution.setPreventedNormalWoundsFromSavesRerolls(savedNormalWoundsFromRerolls);
            savedNormalWounds += savedNormalWoundsFromRerolls;
        }
        resolution.setPreventedNormalWoundsFromSaves(savedNormalWounds);

        // Feel no pain saves
        var preventedWoundsFromFNP = 0d;
        if(feelNoPain.isPresent()) {
            var fnp = feelNoPain.get();
            var fnpRollGoalProbability = fnp.getRollValue() / 6;
            //  Normal
            var remainingNormalWounds = normalWounds - savedNormalWounds;
            var fnpedNormalWounds = 0d;
            if(fnp.getType() == FeelNoPainType.ALL) {
                fnpedNormalWounds += remainingNormalWounds * fnpRollGoalProbability;
                resolution.setPreventedNormalWoundsFromFNP(fnpedNormalWounds);
                savedNormalWounds -= fnpedNormalWounds;
                preventedWoundsFromFNP += fnpedNormalWounds;
            }

            // Mortal wounds
            var fnpedMortalWounds = 0d;
            if(fnp.getType() == FeelNoPainType.ALL || fnp.getType() == FeelNoPainType.MORTAL_WOUNDS) {
                fnpedMortalWounds += mortalWounds * fnpRollGoalProbability;
                resolution.setPreventedMortalWoundsFromFNP(fnpedMortalWounds);
                mortalWounds -= fnpedMortalWounds;
                preventedWoundsFromFNP += fnpedMortalWounds;
            }

            resolution.setPreventedWoundsFromFNP(fnpedNormalWounds + fnpedMortalWounds);
        }

        resolution.setPreventedNormalWounds(savedNormalWounds);
        resolution.setPreventedWounds(savedNormalWounds + preventedWoundsFromFNP);

        return resolution;
    }

    private Optional<Reroll> calculateBestReroll(int saveTarget) {
        if(hasRerollOneNormalSaves) {
            return Optional.of(Reroll.rerollOnOnes());
        } else if (hasRerollFailedNormalSaves || hasRerollAnyNormalSaves) {
            return Optional.of(fishFor6NormalSaves && hasRerollAnyNormalSaves ? Reroll.rerollOnNot(Set.of(6)) :
                    Reroll.rerollFailures(saveTarget));
        } else {
            return Optional.empty();
        }
    }

    private int calculateBestSaveFor(Weapon weapon) {
        var ap = weapon.getAp();
        // Augment basic save with potential modifiers
        var finalSave = normalSave;
        var finalIsInCover = isInCover;
        //   Indirect fire save bonus
        if(weapon.hasIndirectFire() && !isVisible) {
            finalIsInCover = true;
        }
        //   Cover save bonus
        if(finalIsInCover && !weapon.ignoresCover()) {
            // Saves better than +3 get no cover benefits from ap 0 weapons
            if(normalSave > 3 || ap > 0) {
                finalSave++;
            }
        }
        // Subtract AP (a higher save is worse)
        finalSave += weapon.getAp();
        finalSave = Math.min(finalSave, 6);

        if(invulnerableSave.isPresent()) {
            if(invulnerableSave.get().appliesTo(weapon)) {
                var invulnerabilitySave = invulnerableSave.get().getRollValue();
                if(invulnerabilitySave < finalSave) {
                    finalSave = invulnerabilitySave;
                }
            }
        }

        return finalSave;
    }
}
